from .borders import bordered
from .columnify import columnify
from .grow import grow
from .highlight import highlight
from .resize import resize
from .sortify import sortify

__all__ = ["bordered", "columnify", "grow", "highlight", "resize", "sortify"]



from builtins import super

from ..style import Color, Style
from . import ansi16, ansicodes


AIX_COLOR_CODES = {
    ansi16.DARKGRAY: [ansicodes.DARKGRAY],
    ansi16.RED: [ansicodes.BRIGHTRED],
    ansi16.LIME: [ansicodes.BRIGHTGREEN],
    ansi16.YELLOW: [ansicodes.BRIGHTYELLOW],
    ansi16.BLUE: [ansicodes.BRIGHTBLUE],
    ansi16.MAGENTA: [ansicodes.MAGENTA],
    ansi16.CYAN: [ansicodes.BRIGHTCYAN],
    ansi16.WHITE: [ansicodes.WHITE],
}


class AixTerminal(ansi16.Ansi16Terminal):
    def __init__(self):
        super().__init__()
        self.fg_colormap.update(AIX_COLOR_CODES)
        for key in AIX_COLOR_CODES:
            fgcodes = AIX_COLOR_CODES[key]
            bgcodes = [ansicodes.fg_to_bg(code) for code in fgcodes]
            self.bg_colormap[key] = bgcodes

    def configure_style(self, style):
        if style.fg.isreset():
            fg = Color()
        else:
            fg = self._hsl_to_16color(*style.fg.hsl)
        if style.bg.isreset():
            bg = Color()
        else:
            bg = self._hsl_to_16color(*style.bg.hsl)

        def issupported(style):
            return style in self.supported_attrs

        attrs = filter(issupported, style.attrs)
        return Style(fg, bg, attrs)

    def assert_compatible_stylediff(self, style_diff):
        if style_diff.fg is not None:
            self.assert_compatible_color(style_diff.fg)
        if style_diff.bg is not None:
            self.assert_compatible_color(style_diff.bg)
        self.assert_compatible_attrs(style_diff.attrs.keys())

    def test(self):
        # remove the first "No color" color
        colors = self.supported_colors[1:]
        swatches = [
            ["" for x in range(len(colors))] for y in range(len(colors))
        ]

        for i in range(len(colors)):
            for j in range(len(colors)):
                style = Style(fg=colors[i], bg=colors[j])
                diff = style.diff(Style())
                reset_diff = Style().diff(style)
                color = self.encode_sequence(diff)
                reset = self.encode_sequence(reset_diff)
                swatches[i][j] = "{clr} A {reset}".format(
                    clr=color, reset=reset
                )

        text_rows = ["~Lilies~", "Terminal test: ANSI 8 Colors, Extended", ""]
        text_rows.append("Color table:")
        color_rows = map(lambda rowlist: "".join(rowlist), swatches)

        print("\n".join(text_rows + list(color_rows)))
        self.test_attributes()
        print("\n")



from colorama import init
from colorama.initialise import deinit, reinit


class Win32Mixin(object):
    def setup(self, appstate):
        if appstate.get("colorama_init"):
            reinit()
            return
        init()
        appstate["colorama_init"] = True

    def teardown(self, appstate):
        deinit()



from .api import columnify, bordered, grow, highlight, resize, sortify
from .base_utils import wilt
from .manage import lilies_init
from . import compiler

__version__ = "0.3.2"

version = VERSION = __version__


def print_test():
    term = compiler.get_compiler().term
    term.test()


lilies_init()

__all__ = [
    # helpers
    "grow",
    "wilt",
    # actions
    "bordered",
    "columnify",
    "highlight",
    "resize",
    "sortify",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import unittest
from lilies import wilt, resize
from lilies.objects.lilyblock import LilyBlock


class TestResize(unittest.TestCase):
    def setUp(self):
        self.strings = [
            "hello",
            "dfDfEEFdfaC",
            "Mister John",
            "mr. jOhn",
            "iSn't it",
            "comma,separated,values",
            "trailing,comma,",
            ",",
        ]
        self.block_str = os.linesep.join(self.strings)
        self.twenty_chars = "12345678901234567890"

    def test_empty_align_fails(self):
        with self.assertRaises(TypeError):
            resize(self.twenty_chars, 20, align="")

    def test_garbage_align_fails(self):
        with self.assertRaises(TypeError):
            resize(self.twenty_chars, 20, align="garbage")

    def test_garbage_x_align_fails(self):
        with self.assertRaises(TypeError):
            resize(self.twenty_chars, align="top garbage")

    def test_garbage_y_align_fails(self):
        with self.assertRaises(TypeError):
            resize(self.twenty_chars, align="garbage left")

    def test_resize_vertically_respects_fill_character(self):
        resized = resize("   ", 5, 3, "center", character="a")
        expected = "\n".join(["aaaaa", "a   a", "aaaaa"])
        self.assertEqual(expected, wilt(resized))

    def test_resize_to_same_size_leaves_unchanged(self):
        resized = resize(self.twenty_chars, 20, add_elipsis=False)
        self.assertEqual(self.twenty_chars, wilt(resized))

    def test_resize_same_width_with_elipsis_leaves_unchanged(self):
        resized = resize(self.twenty_chars, 20, add_elipsis=True)
        self.assertEqual(self.twenty_chars, wilt(resized))

    def test_resize_same_width_centered_leaves_unchanged(self):
        resized = resize(self.twenty_chars, 20, align="center")
        self.assertEqual(self.twenty_chars, wilt(resized))

    def test_resize_smaller_left_no_elipsis_truncates(self):
        expected = self.twenty_chars[:10]
        resized = resize(
            self.twenty_chars, 10, align="left", add_elipsis=False
        )
        self.assertEqual(expected, wilt(resized))

    def test_resize_smaller_left_with_elipsis_truncates(self):
        expected = self.twenty_chars[:8] + ".."
        resized = resize(self.twenty_chars, 10, align="left", add_elipsis=True)
        self.assertEqual(expected, wilt(resized))

    def test_resize_smaller_right_with_elipsis_truncates(self):
        # we basically expect this to be the same. Don't truncate left side.
        expected = self.twenty_chars[:8] + ".."
        resized = resize(
            self.twenty_chars, 10, align="right", add_elipsis=True
        )
        self.assertEqual(expected, wilt(resized))

    def test_resize_smaller_center_with_elipsis_truncates(self):
        # we basically expect this to be the same. Don't truncate left side.
        expected = self.twenty_chars[:8] + ".."
        resized = resize(
            self.twenty_chars, 10, align="center", add_elipsis=True
        )
        self.assertEqual(expected, wilt(resized))

    def test_resize_horiz_larger_left(self):
        expected = self.twenty_chars + " " * 20
        resized = resize(self.twenty_chars, 40, align="left")
        self.assertEqual(expected, wilt(resized))

    def test_resize_horiz_larger_right(self):
        expected = " " * 20 + self.twenty_chars
        resized = resize(self.twenty_chars, 40, align="right")
        self.assertEqual(expected, wilt(resized))

    def test_resize_horiz_larger_center_even(self):
        expected = " " * 10 + self.twenty_chars + " " * 10
        resized = resize(self.twenty_chars, 40, align="center")
        self.assertEqual(expected, wilt(resized))

    def test_resize_horiz_larger_center_odd(self):
        expected = " " * 9 + self.twenty_chars + " " * 10
        resized = resize(self.twenty_chars, 39, align="center")
        self.assertEqual(expected, wilt(resized))

    def test_resize_block_horiz_larger(self):
        control_group = [
            "hello                  ",
            "dfDfEEFdfaC            ",
            "Mister John            ",
            "mr. jOhn               ",
            "iSn't it               ",
            "comma,separated,values ",
            "trailing,comma,        ",
            ",                      ",
        ]
        control = os.linesep.join(control_group)
        block = LilyBlock(self.block_str)
        result = resize(block, width=23)
        self.assertEqual(control, wilt(result))

    def test_resize_block_vert_top_larger(self):
        control_group = [
            "hello                 ",
            "dfDfEEFdfaC           ",
            "Mister John           ",
            "mr. jOhn              ",
            "iSn't it              ",
            "comma,separated,values",
            "trailing,comma,       ",
            ",                     ",
            "                      ",
            "                      ",
            "                      ",
        ]
        control = os.linesep.join(control_group)
        block = LilyBlock(self.block_str)
        result = resize(block, height=11, align="top")
        self.assertEqual(control, wilt(result))

    def test_resize_block_vert_bottom_larger(self):
        control_group = [
            "                      ",
            "                      ",
            "hello                 ",
            "dfDfEEFdfaC           ",
            "Mister John           ",
            "mr. jOhn              ",
            "iSn't it              ",
            "comma,separated,values",
            "trailing,comma,       ",
            ",                     ",
        ]
        control = os.linesep.join(control_group)
        block = LilyBlock(self.block_str)
        result = resize(block, height=10, align="bottom")
        self.assertEqual(control, wilt(result))

    def test_resize_block_vert_center_larger(self):
        control_group = [
            "                      ",
            "                      ",
            "hello                 ",
            "dfDfEEFdfaC           ",
            "Mister John           ",
            "mr. jOhn              ",
            "iSn't it              ",
            "comma,separated,values",
            "trailing,comma,       ",
            ",                     ",
            "                      ",
            "                      ",
            "                      ",
        ]
        control = os.linesep.join(control_group)
        block = LilyBlock(self.block_str)
        result = resize(block, height=13, align="center left")
        self.assertEqual(control, wilt(result))

    def test_resize_block_vert_top_smaller(self):
        control_group = [
            "hello      ",
            "dfDfEEFdfaC",
            "Mister John",
            "mr. jOhn   ",
            "iSn't it   ",
        ]
        control = os.linesep.join(control_group)
        block = LilyBlock(self.block_str)
        result = resize(block, height=5, align="top")
        self.assertEqual(control, wilt(result))

    def test_resize_block_vert_bottom_smaller(self):
        control_group = [
            "dfDfEEFdfaC           ",
            "Mister John           ",
            "mr. jOhn              ",
            "iSn't it              ",
            "comma,separated,values",
            "trailing,comma,       ",
            ",                     ",
        ]
        control = os.linesep.join(control_group)
        block = LilyBlock(self.block_str)
        result = resize(block, height=7, align="bottom")
        self.assertEqual(control, wilt(result))

    def test_resize_block_vert_center_smaller(self):
        control_group = [
            "dfDfEEFdfaC           ",
            "Mister John           ",
            "mr. jOhn              ",
            "iSn't it              ",
            "comma,separated,values",
        ]
        control = os.linesep.join(control_group)
        block = LilyBlock(self.block_str)
        result = resize(block, height=5, align="center left")
        self.assertEqual(control, wilt(result))



from .detect import detect_terminal, print_features
from .nocolor import NoColorTerminal
from .ansi8 import Ansi8Terminal
from .ansi8open import Ansi8OpenTerminal
from .aixterm import AixTerminal
from .ansi16 import Ansi16Terminal
from .ansi256 import Ansi256Terminal
from .truecolor import TrueColorTerminal
from .windows import WinTrueColor, WinLegacy, Win256Color

__all__ = [
    "detect_terminal",
    "print_features",
    "NoColorTerminal",
    "Ansi8Terminal",
    "Ansi8OpenTerminal",
    "AixTerminal",
    "Ansi16Terminal",
    "Ansi256Terminal",
    "TrueColorTerminal",
    "WinTrueColor",
    "WinLegacy",
    "Win256Color",
]



from __future__ import print_function
from builtins import int, super

from ..style import Color, Style
from ..style.colors import hsl_to_rgb
from .exceptions import UnsupportedColorException
from .ansi8 import Ansi8Terminal
from .attr_mixin import AnsiAttributesMixin
from . import ansicodes


class TrueColorTerminal(Ansi8Terminal, AnsiAttributesMixin):
    """
    An ANSI-sequence based truecolor terminal
    """

    def __init__(self):
        super().__init__()
        self.supported_attrs = ["bold", "dim", "underline", "blink"]

    def configure_style(self, style):
        def issupported(style):
            return style in self.supported_attrs

        attrs = [attr for attr in style.attrs if issupported(attr)]
        return Style(style.fg, style.bg, attrs)

    def assert_compatible_color(self, color):
        if color == Color():
            return
        for val in color.rgb:
            if not isinstance(val, int):
                msg = "color must have integer rgb values: rgb{clr}".format(
                    clr=color.rgb
                )
                raise UnsupportedColorException(msg)

    def _build_fg_codes(self, fg_color):
        if fg_color is None:
            return []
        if fg_color == Color():
            return [ansicodes.NOCOLOR]
        (r, g, b) = fg_color.rgb
        formatted = "38;2;{r};{g};{b}".format(r=r, g=g, b=b)
        return [formatted]

    def _build_bg_codes(self, bg_color):
        if bg_color is None:
            return []
        if bg_color == Color():
            return [ansicodes.fg_to_bg(ansicodes.NOCOLOR)]
        (r, g, b) = bg_color.rgb
        formatted = "48;2;{r};{g};{b}".format(r=r, g=g, b=b)
        return [formatted]

    def test(self):
        def _color_cell(h, s, l):
            r, g, b = hsl_to_rgb(h, s, l)
            return "\033[48;2;{r};{g};{b}m ".format(r=r, g=g, b=b)

        width = 75
        height = 24
        rows = ["~Lilies~", "Terminal test: ANSI True Color", ""]
        rows.append("Colors sample:")
        for y in range(height):
            lightness = float(height - y) / float(height) * 100
            row = []
            for x in range(width):
                hue = float(x) / float(width) * 360
                row.append(_color_cell(hue, 100, lightness))
            rows.append("".join(row) + "\033[0m")
        rows.append("")
        rows.append("Grayscale sample:")
        rows.append(
            "".join(
                [
                    _color_cell(0, 0, float(x) / float(width) * 50)
                    for x in range(width)
                ]
            )
            + "\033[0m"
        )
        rows.append(
            "".join(
                [
                    _color_cell(0, 0, 100 - float(x) / float(width) * 50)
                    for x in range(width)
                ]
            )
            + "\033[0m"
        )
        rows.append("")
        print("\n".join(rows))
        self.test_attributes()



##########################################################################
# Lilies
# By: Matt Zychowski (copyright 2014-2021)
#
# A library wrapped around colorama for using colored strings in the terminal
# window.  Provides advanced manipulation of colored strings, including
# accurate slicing.
#
##########################################################################

from __future__ import print_function
from builtins import str
from . import grow, __version__


print(grow("Lilies!", "green"))
print(
    grow("A test formatting and coloring tool for the command line", "yellow")
)
print("Version: " + str(__version__))
print("Author: Matt Zychowski")
print()
print("=" * 50)
print()



import unittest
from lilies import highlight
from lilies.objects.lilyblock import LilyBlock
from lilies.objects.lilystring import LilyString
from lilies.style import parse_style, Style


class TestHighlight(unittest.TestCase):
    def setUp(self):
        self.hello_world_flat = "Hello, world!"
        self.hello_world_block = LilyBlock("Hello, \nworld!")
        self.l_locations = [2, 3, 10]
        self.highlighted_style = parse_style("yellow")

    def test_highlights_flat_string(self):
        highlighted = highlight(self.hello_world_flat, "l")
        for i in range(len(highlighted)):
            char_style = highlighted.style_at(i)
            if i in self.l_locations:
                self.assertEqual(char_style, self.highlighted_style)
            else:
                self.assertEqual(char_style, Style())

    def test_highlighting_flat_string_returns_lilystring(self):
        highlighted = highlight(self.hello_world_flat, "l")
        self.assertEqual(type(highlighted), LilyString)

    def test_highlights_block(self):
        highlighted = highlight(self.hello_world_block, "l")
        char_counter = 0
        for row in highlighted:
            for i in range(len(row)):
                char_style = row.style_at(i)
                if char_counter in self.l_locations:
                    self.assertEqual(char_style, self.highlighted_style)
                else:
                    self.assertEqual(char_style, Style())
                char_counter += 1

    def test_highlighting_block_returns_block(self):
        highlighted = highlight(self.hello_world_block, "l")
        self.assertEqual(type(highlighted), LilyBlock)

    def test_highlighting_one_char(self):
        locations = self.l_locations[:1]
        highlighted = highlight(self.hello_world_flat, "l", num=1)
        for i in range(len(highlighted)):
            char_style = highlighted.style_at(i)
            if i in locations:
                self.assertEqual(char_style, self.highlighted_style)
            else:
                self.assertEqual(char_style, Style())

    def test_highlighting_two_chars(self):
        locations = self.l_locations[:2]
        highlighted = highlight(self.hello_world_flat, "l", num=2)
        for i in range(len(highlighted)):
            char_style = highlighted.style_at(i)
            if i in locations:
                self.assertEqual(char_style, self.highlighted_style)
            else:
                self.assertEqual(char_style, Style())



from .base import BaseTerminal
from .exceptions import UnsupportedAttributeException
from .exceptions import UnsupportedColorException
from ..style import Color, Style


class NoColorTerminal(BaseTerminal):
    """
    Base terminal with no colors at all
    """

    def __init__(self):
        self.supported_attrs = []
        self.supported_colors = [Color()]

    def setup(self, appstate):
        pass

    def teardown(self, appstate):
        pass

    def configure_style(self, style):
        # "any color you like"
        return Style()

    def assert_compatible_color(self, color):
        if color not in self.supported_colors:
            msg = "color unsupported in this terminal: rgb{clr}".format(
                clr=color.rgb
            )
            raise UnsupportedColorException(msg)

    def assert_compatible_attrs(self, attrs):
        for attr in attrs:
            if attr not in self.supported_attrs:
                msg = "attr unsupported in this terminal: '{attr}'".format(
                    attr=attr
                )
                raise UnsupportedAttributeException(msg)

    def _build_fg_codes(self, fg_color):
        return []

    def _build_bg_codes(self, bg_color):
        return []

    def _build_attr_codes(self, attrs):
        return []

    def _build_reset_sequence(self):
        return []

    def test(self):
        lines = ["~Lilies~", "Terminal test: No Color"]
        lines.append("")
        lines.append("All attributes disabled, nothing to show!")
        print("\n".join(lines))



from .grow import grow
from .resize import resize
from .sortify import sortify
from ..objects.lilyblock import block


class ColumnBuilder:
    def __init__(
        self, column_count=0, total_width=80, align="left", min_spacing=3
    ):
        self.column_count = column_count
        self.total_width = total_width
        self.align = align
        self.min_spacing = min_spacing

    def make_columns(self, iterable, sort=True, horizontal=False):
        if not iterable:
            return block("")

        iterable = [grow(item) for item in iterable]
        column_count = self._get_column_count(iterable)
        entries_per_column, remainder = divmod(len(iterable), column_count)
        if remainder > 0:
            entries_per_column += 1

        if sort:
            iterable = sortify(iterable)
        iterable = self._resize_all(iterable, column_count)

        if horizontal:
            rows = self._make_rows_horizontally(
                iterable, column_count, entries_per_column
            )
        else:
            rows = self._make_rows_vertically(
                iterable, column_count, entries_per_column
            )

        return self._merge_rows_to_lilyblock(rows)

    def _make_rows_horizontally(
        self, iterable, column_count, entries_per_column
    ):
        rows = []
        for i in range(entries_per_column):
            start = column_count * i
            stop = start + column_count
            rows.append(iterable[start:stop])
        return rows

    def _make_rows_vertically(
        self, iterable, column_count, entries_per_column
    ):
        rows = []
        for i in range(entries_per_column):
            rows.append(iterable[i::entries_per_column])
        return rows

    def _get_column_count(self, iterable):
        if self.column_count != 0:
            return self.column_count

        longest_item_size = max([len(item) for item in iterable])
        columns = self.total_width // (longest_item_size + self.min_spacing)

        # If we have fewer items than columns, we aren't going to
        # fill all of the columns.
        if len(iterable) < columns:
            columns = len(iterable)
        if columns == 0:
            return 1
        return columns

    def _resize_all(self, iterable, column_count):
        total_spacing = self.min_spacing * (column_count - 1)
        usable_width = self.total_width - total_spacing
        max_cell_size = usable_width // column_count
        max_item_size = max([len(item) for item in iterable])
        item_size = min([max_cell_size, max_item_size])
        return [resize(s, item_size, align=self.align) for s in iterable]

    def _merge_rows_to_lilyblock(self, rows):
        space = grow(" " * self.min_spacing)
        rows = [space.join(row) for row in rows]
        return block(rows)


def columnify(
    iterable,
    columns=0,
    width=80,
    align="left",
    sort=True,
    min_spacing=3,
    horizontal=False,
):
    builder = ColumnBuilder(
        column_count=columns,
        total_width=width,
        align=align,
        min_spacing=min_spacing,
    )
    return builder.make_columns(iterable, sort=sort, horizontal=horizontal)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
import os
from lilies import columnify, wilt


class TestBordered(unittest.TestCase):
    def setUp(self):
        self.console_width = 30
        self.columns_10_entries_vertical = os.linesep.join(
            [
                "01abcde   05abcde   09abcde",
                "02abcde   06abcde   10abcde",
                "03abcde   07abcde",
                "04abcde   08abcde",
            ]
        )
        self.columns_10_entries_horizontal = os.linesep.join(
            [
                "01abcde   02abcde   03abcde",
                "04abcde   05abcde   06abcde",
                "07abcde   08abcde   09abcde",
                "10abcde",
            ]
        )
        self.one_column_long_entries = os.linesep.join(
            [
                "abcdefghijklmnopqrstuvwxyz0123",
                "0" * 25 + " " * 5,
                "9" * 28 + " " * 2,
            ]
        )
        self.one_entry = "abcd"
        self.empty = ""

    def test_columnify_works_with_10_entries_vertical(self):
        expected = self.columns_10_entries_vertical

        entries = expected.split()
        actual = wilt(columnify(entries, width=self.console_width))
        self.assertEqual(expected, actual)

    def test_columnify_works_with_10_entries_horizontal(self):
        expected = self.columns_10_entries_horizontal

        entries = expected.split()
        actual = wilt(
            columnify(entries, width=self.console_width, horizontal=True)
        )
        self.assertEqual(expected, actual)

    def test_columnify_works_with_long_entries(self):
        expected = self.one_column_long_entries

        entries = expected.split()
        actual = wilt(columnify(entries, width=self.console_width, sort=False))
        self.assertEqual(expected, actual)

    def test_columnify_works_with_one_entry(self):
        expected = self.one_entry

        entries = expected.split()
        actual = wilt(columnify(entries, width=self.console_width))
        self.assertEqual(expected, actual)

    def test_columnify_works_with_empty_list(self):
        expected = self.empty

        entries = expected.split()
        actual = wilt(columnify(entries, width=self.console_width))
        self.assertEqual(expected, actual)



from ..base_utils import wilt


def sortify(iter, case_insensitive=True):
    if case_insensitive:

        def sortkey(s):
            return wilt(s).lower()

    else:
        sortkey = wilt
    return sorted(list(iter), key=sortkey)



import os
from copy import deepcopy
from builtins import map
from builtins import str
from future.moves.itertools import zip_longest

from ..base_utils import isstringish, islilyblock
from ..compiler import compile_all
from ..style import Style
from .base import LilyBase
from .lilystring import lstr


def block(obj, *args, **kwargs):
    if islilyblock(obj):
        return obj
    return LilyBlock(obj, *args, **kwargs)


class LilyBlock(LilyBase):
    def __init__(self, rows=[], style=Style(), newline_char=os.linesep):
        if isstringish(rows):
            rows = rows.split(newline_char)
        if islilyblock(rows):
            rows = rows._copy_rows()
        if not hasattr(rows, "__iter__"):
            rows = [rows]

        self._endl = newline_char
        grower = lstr if style.is_default() else lambda s: lstr(s, style)
        rows = list(map(grower, rows))

        split_rows = []
        for row in rows:
            split_rows += row.split(self._endl)
        self._rows = split_rows

    def __str__(self):
        return self._endl.join(map(str, self._rows))

    def __repr__(self):
        return "c'''" + self.__str__() + "'''"

    def __getitem__(self, key):
        if not (isinstance(key, slice) or isinstance(key, int)):
            raise TypeError("Invalid argument type, looking for int or slice")

        sliced = self._copy_rows().__getitem__(key)
        return LilyBlock(sliced)

    def __iter__(self):
        for row in self._rows:
            yield row

    def __add__(self, other):
        return self.concat(other)

    def __radd__(self, other):
        try:
            return other.__add__(self)
        except TypeError:
            return LilyBlock(other).__add__(self)

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError("can't multiply sequence by non-int")
        if other < 1:
            return LilyBlock()
        new = LilyBlock(self._copy_rows())
        for _ in range(other - 1):
            new = new.concat(self)
        return new

    def __rmul__(self, other):
        return self.__mul__(other)

    def __iadd__(self, other):
        return self.__add__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __gt__(self, other):
        return self.wilt() > other

    def __lt__(self, other):
        return self.wilt() < other

    def __ge__(self, other):
        return self.wilt() >= other

    def __le__(self, other):
        return self.wilt() <= other

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return sum(map(len, self._rows))

    def __hash__(self):
        with compile_all:
            return hash(self.__str__())

    def _copy_rows(self):
        return deepcopy(self._rows)

    def _isstringish(self):
        return False

    def _isblockish(self):
        return True

    def wilt(self):
        rows = map(lambda r: r.wilt(), self._rows)
        return self._endl.join(rows)

    def append(self, stringlike, align="left"):
        other = block(stringlike)
        if align == "right":
            return self._append_right(other)
        if align == "center":
            return self._append_center(other)
        top_rows = self._copy_rows()
        bottom_rows = other._copy_rows()

        rows = [lstr(row) for row in top_rows + bottom_rows]
        return LilyBlock(rows)

    def _append_right(self, other):
        if other.width > self.width:
            top_rows = [row.rjust(other.width) for row in self._copy_rows()]
            bottom_rows = other._copy_rows()
        else:
            top_rows = self._copy_rows()
            bottom_rows = [row.rjust(self.width) for row in other._copy_rows()]
        rows = [lstr(row) for row in top_rows + bottom_rows]
        return LilyBlock(rows)

    def _append_center(self, other):
        if other.width > self.width:
            top_rows = [row.center(other.width) for row in self._copy_rows()]
            bottom_rows = other._copy_rows()
        else:
            top_rows = self._copy_rows()
            bottom_rows = [
                row.center(self.width) for row in other._copy_rows()
            ]
        rows = [lstr(row) for row in top_rows + bottom_rows]
        return LilyBlock(rows)

    def concat(self, lilyblock, squash=False):
        rows = self._copy_rows()
        if not squash:
            width = self.width
            rows = [s.ljust(width) for s in rows]
        other_rows = lilyblock._copy_rows()
        zipped = zip_longest(rows, other_rows, fillvalue=lstr(""))
        new_rows = [r[0] + r[1] for r in zipped]
        return LilyBlock(new_rows)

    @property
    def width(self):
        return max([len(row) for row in self._rows])

    @property
    def height(self):
        return len(self._rows)

    def _map_rebuild(self, func):
        rows = self._copy_rows()
        mapped = map(func, rows)
        return LilyBlock(mapped)

    def _vstrip_by_iter(self, iter, chars):
        strip_complete = False
        for row in iter:
            if strip_complete:
                yield row
                continue
            stripped = row.strip(chars)
            if len(stripped) != 0:
                strip_complete = True
                yield row

    def tstrip(self, chars=None):
        rows = self._vstrip_by_iter(self._copy_rows(), chars)
        return LilyBlock(rows)

    def bstrip(self, chars=None):
        rows = self._vstrip_by_iter(self._copy_rows()[::-1], chars)
        return LilyBlock(list(rows)[::-1])

    def lstrip(self, chars=None):
        def func(s):
            return s.lstrip(chars)

        return self._map_rebuild(func)

    def rstrip(self, chars=None):
        def func(s):
            return s.rstrip(chars)

        return self._map_rebuild(func)

    def strip_x(self, chars=None):
        def func(s):
            return s.strip(chars)

        return self._map_rebuild(func)

    def strip_y(self, chars=None):
        return self.tstrip(chars).bstrip(chars)

    def strip(self, chars=None, x=True, y=True):
        stripped = self.strip_x(chars) if x else self
        return stripped.strip_y(chars) if y else LilyBlock(self._copy_rows())



from ..style import Style


class LilyStringCompiler(object):
    """
    A compiler object that will take a series of
    LilyStringPieces and compile them to display
    in a particular terminal. Designed to 'just work'
    even when the colors/styles used are not fully
    supported within the terminal environment chosen.
    """

    def __init__(self, terminal):
        self.term = terminal

    def compile(self, pieces):
        compiled_parts = []
        previous_style = Style()
        for piece in pieces:
            cur_style = self.term.configure_style(piece.style)
            diff = cur_style.diff(previous_style)
            previous_style = cur_style
            compiled_parts.append(self.term.encode_sequence(diff))
            compiled_parts.append(piece.text)
        # reset
        diff = Style().diff(previous_style)
        compiled_parts.append(self.term.encode_sequence(diff))
        return "".join(compiled_parts)



import os
import sys
import platform
import subprocess
from builtins import object

from .nocolor import NoColorTerminal
from .ansi8 import Ansi8Terminal
from .ansi8open import Ansi8OpenTerminal
from .aixterm import AixTerminal
from .ansi16 import Ansi16Terminal
from .ansi256 import Ansi256Terminal
from .truecolor import TrueColorTerminal
from .windows import WinTrueColor, Win256Color, WinLegacy


# Note: The order, names, and numbers
# on these may change at any time.
# Do not rely on them long-term.
NO_COLOR = 1
ANSI8 = 2
ANSI8_OPEN = 3
ANSI16 = 4
AIXTERM = 5
ANSI256 = 6
FULL_RGB = 7
WIN10_24BIT = 8
WIN10_256 = 9
WIN32 = 10

TERM_CLASSES = {
    NO_COLOR: NoColorTerminal,
    ANSI8: Ansi8Terminal,
    ANSI8_OPEN: Ansi8OpenTerminal,
    ANSI16: Ansi16Terminal,
    AIXTERM: AixTerminal,
    ANSI256: Ansi256Terminal,
    FULL_RGB: TrueColorTerminal,
    WIN10_24BIT: WinTrueColor,
    WIN10_256: Win256Color,
    WIN32: WinLegacy,
}


_term = None


class SystemProperties(object):
    def __init__(
        self,
        pltform="",
        env={},
        stdout=None,
        version=["0", "0", "0"],
        tput_colors=None,
    ):
        self.platform = pltform
        self.env = env
        self.stdout = stdout
        self.ver = version
        self.tput_colors = tput_colors


def detect_terminal():
    global _term
    if _term is not None:
        return _term

    term_type = _detect()
    term_cls = TERM_CLASSES.get(term_type) or NoColorTerminal
    _term = term_cls()
    return _term


def _get_system():
    plat = platform.system()
    env = os.environ
    stdout = sys.stdout
    if "Windows" in plat:
        version = platform.version().split(".")
    else:
        version = platform.release().split(".")
    try:
        result = subprocess.check_output(["tput", "colors"])
        tput_colors = int(result.strip())
    except Exception:
        tput_colors = None

    return SystemProperties(plat, env, stdout, version, tput_colors)


def print_features(system=_get_system()):
    print("Platform: {}".format(system.platform))
    print("Version: {}".format(system.ver))
    print("isatty: {}".format(system.stdout.isatty()))
    print("NOCOLOR: {}".format(system.env.get("NO_COLOR") is not None))
    print("TERM: {}".format(system.env.get("TERM")))
    print("TERM_PROGRAM: {}".format(system.env.get("TERM_PROGRAM")))
    print("COLORTERM: {}".format(system.env.get("COLORTERM")))
    print("TPut Colors: {}".format(system.tput_colors))


def _detect(system=_get_system()):
    if not system.stdout.isatty() or system.env.get("NO_COLOR") is not None:
        return NO_COLOR

    # Check colorterm env for modern
    # terminal settings
    colorterm = system.env.get("COLORTERM")
    if colorterm:
        if colorterm in {"truecolor", "24bit"}:
            return FULL_RGB
        if colorterm in {"8bit"}:
            return ANSI256

    termprogram = system.env.get("TERM_PROGRAM")
    if termprogram:
        if termprogram in {"iTerm.app", "Hyper"}:
            return FULL_RGB
        if termprogram in {"Apple_Terminal"}:
            return ANSI256

    if "CYGWIN" in system.platform:
        return _get_cygwin_color(system)

    term = system.env.get("TERM")
    if term:
        if "screen-256" in term or "xterm-256" in term:
            return ANSI256
        if "vt100" in term:
            return ANSI16
        if term in {"screen", "xterm", "color", "ansi", "linux"}:
            return ANSI8_OPEN

    if "Windows" in system.platform:
        return _get_windows_color(system)

    if term:
        if "aixterm" in term or "xterm-16" in term:
            return AIXTERM

        if term in {"dumb"}:
            return NO_COLOR

    termcolors = _get_termcolors(system.tput_colors)
    if termcolors:
        return termcolors

    # fuck it, most terms should be 8 color...
    return ANSI8


def _get_cygwin_color(system):
    # Cygwin takes over the system platform variable.
    # It's hard to determine the exact version that
    # 24-bit color was introduced, but anyone with a
    # version older than 2.9 should update...
    try:
        version_major = int(system.ver[0])
        version_minor = int(system.ver[1])
        if version_major >= 3:
            return FULL_RGB
        if version_major == 2 and version_minor >= 9:
            return FULL_RGB
    except Exception:
        pass
    return AIXTERM


def _get_windows_color(system):
    term = system.env.get("TERM")
    mingw = system.env.get("MINGW64")

    # detects git bash, which needs colorama.
    if term and "cygwin" in term and mingw and "MSYSTEM" in mingw:
        return WIN32

    release = system.ver
    if release[0] == "10":
        # Windows10 started to add real color support.
        # Detect if this version is up to date
        micro_ver = int(release[2])
        if micro_ver > 10586:
            return WIN10_24BIT if micro_ver > 14931 else WIN10_256

    return WIN32


def _get_termcolors(tput_colors):
    if tput_colors is None:
        return None
    if tput_colors >= 256:
        # Honestly, if this is more than 256,
        # we can't be sure if it's truecolor.
        # There are some weird terminals out there,
        # including 4352 color support?
        return ANSI256
    if tput_colors >= 16:
        # sadly I can't be certain that this will
        # properly support bright colors?
        return ANSI8_OPEN
    if tput_colors >= 8:
        return ANSI8
    return NO_COLOR



from builtins import super

from .ansi8 import Ansi8Terminal
from .attr_mixin import AnsiAttributesMixin
from . import ansicodes
from .exceptions import UnsupportedColorException
from ..style import Color, Style
from ..style.palette import (
    ORDERED_ANSI_16_COLOR_PALETTE,
    ANSI_16_COLOR_PALETTE,
)

# ANSI 256 color uses 5 color intensities plus "none" for each of R, G, B.
# This identifies the standard "color space" for colored text, not including
# an additional grayscale band.
COLOR_MAGNITUDES = [95, 135, 175, 215, 255]
COLOR_INTENSITY_MAP = {
    0: 0,
    COLOR_MAGNITUDES[0]: 1,
    COLOR_MAGNITUDES[1]: 2,
    COLOR_MAGNITUDES[2]: 3,
    COLOR_MAGNITUDES[3]: 4,
    COLOR_MAGNITUDES[4]: 5,
}


# Does not include white or black, which are special cases.
# This tops out at 238, then the next magnitude full white.
GRAY_MAGNITUDES = range(8, 240, 10)
WHITE = Color(rgb=(255, 255, 255))


class Ansi256Terminal(AnsiAttributesMixin, Ansi8Terminal):
    """
    A 256-color ansi terminal
    """

    def __init__(self):
        super().__init__()
        self.supported_attrs = ["bold", "dim", "italic", "underline", "blink"]

    def configure_style(self, style):
        """
        Configure the style for this terminal
        """
        if style.fg.isreset():
            fg = Color()
        else:
            if style.fg in ANSI_16_COLOR_PALETTE:
                fg = style.fg
            else:
                fg = self._rgb_to_color(*style.fg.rgb)
        if style.bg.isreset():
            bg = Color()
        else:
            if style.bg in ANSI_16_COLOR_PALETTE:
                bg = style.bg
            else:
                bg = self._rgb_to_color(*style.bg.rgb)

        def issupported(style):
            return style in self.supported_attrs

        attrs = filter(issupported, style.attrs)
        attrs = [attr for attr in attrs]
        return Style(fg, bg, attrs)

    def assert_compatible_color(self, color):
        if color == Color():
            return
        (r, g, b) = color.rgb

        # The 256 color space also has the "legacy"
        # original 16 ansi colors. These don't really
        # fit in the other 240 color 'cube'
        if color in ANSI_16_COLOR_PALETTE:
            return

        if r == g and g == b:
            if r in GRAY_MAGNITUDES:
                return
            msg = "color unsupported in this terminal: rgb{clr}".format(
                clr=color.rgb
            )
            raise UnsupportedColorException(msg)

        for color_part in color.rgb:
            if color_part == 0 or color_part in COLOR_MAGNITUDES:
                continue
            msg = "color unsupported in this terminal: rgb{clr}".format(
                clr=color.rgb
            )
            raise UnsupportedColorException(msg)

    def _build_fg_codes(self, fg_color):
        if fg_color is None:
            return []
        if fg_color == Color():
            return [ansicodes.NOCOLOR]
        ansi = self._ansi256_code(fg_color)
        formatted = "38;5;{ansi}".format(ansi=ansi)
        return [formatted]

    def _build_bg_codes(self, bg_color):
        if bg_color is None:
            return []
        if bg_color == Color():
            return [ansicodes.fg_to_bg(ansicodes.NOCOLOR)]
        ansi = self._ansi256_code(bg_color)
        formatted = "48;5;{ansi}".format(ansi=ansi)
        return [formatted]

    def _16color_code(self, color):
        return ORDERED_ANSI_16_COLOR_PALETTE.index(color)

    def _ansi256_code(self, color):
        if color in ANSI_16_COLOR_PALETTE:
            if color == WHITE:
                return 231
            return self._16color_code(color)

        (r, g, b) = color.rgb

        # grayscale
        #
        # We are ignoring the color-cube grays, since they aren't super
        # useful and special casing would be weird.
        if r == g and g == b:
            intensity_ratio = (r - 8) / 247.0
            # 24 different grayscale colors
            return int(round(intensity_ratio * 24)) + 232

        # colors!
        # break back out into a 0-5 intensity
        # then overlay that result over the color cube
        try:
            r_component = COLOR_INTENSITY_MAP[r]
            g_component = COLOR_INTENSITY_MAP[g]
            b_component = COLOR_INTENSITY_MAP[b]
        except KeyError:
            msg = "unexpected color for ANSI256: rgb({r}, {g}, {b})".format(
                r=r, g=g, b=b
            )
            raise UnsupportedColorException(msg)

        return 36 * r_component + 6 * g_component + b_component + 16

    def _rgb_to_color(self, r, g, b):
        if r == g and g == b:
            # Full white/black aren't in the "graybyte" list
            if r > 246:
                return Color(rgb=(255, 255, 255))
            if r < 5:
                return Color(rgb=(0, 0, 0))
            new_c = self._round_graybyte(r)
            return Color(rgb=(new_c, new_c, new_c))
        # Instead of doing a "dumb round" which assumes that the colors are
        # evenly spaced, do a smarter round based on actual outcomes:
        new_r = self._round_colorbyte(r)
        new_g = self._round_colorbyte(g)
        new_b = self._round_colorbyte(b)

        # In rare cases, this winds up being gray, and messing us up.
        # Go back to looking at gray colors.
        if new_r == new_g and new_g == new_b:
            avg_rgb = round((r + g + b) / 3)
            return self._rgb_to_color(avg_rgb, avg_rgb, avg_rgb)

        return Color(rgb=(new_r, new_g, new_b))

    def _round_colorbyte(self, colorbyte):
        # This is kind of arbitrary, but we want to split the color space
        # between 0 and 95 a bit unevenly to try to get slightly more
        # true colors out.
        if colorbyte < 32:
            return 0
        return min(COLOR_MAGNITUDES, key=lambda mag: abs(mag - colorbyte))

    def _round_graybyte(self, graybyte):
        return min(GRAY_MAGNITUDES, key=lambda mag: abs(mag - graybyte))

    def test(self):
        def _get_swatch(indices):
            colors = []
            for i in indices:
                colors.append("\033[48;5;{i}m    ".format(i=i))
            colors.append("\033[0m")
            return "".join(colors)

        lines = ["~Lilies~", "Terminal test: ANSI 256 Colors", ""]

        lines.append("Standard 16 Colors:")
        lines.append("    {sw}".format(sw=_get_swatch(range(8))))
        lines.append("    {sw}".format(sw=_get_swatch(range(8, 16))))
        lines.append("")

        lines.append("Color cube:")
        for i_1, i_2 in zip(
            range(16, 200, 36 * 2), range(16 + 36, 200, 36 * 2)
        ):
            for j in range(6):
                start1 = i_1 + j * 6
                start2 = i_2 + j * 6
                swatch1 = _get_swatch(range(start1, start1 + 6))
                swatch2 = _get_swatch(range(start2, start2 + 6))
                lines.append("    ".join(["", swatch1, swatch2]))
            lines.append("")

        lines.append("Grays:")
        swatch1 = _get_swatch(range(232, 232 + 6))
        swatch2 = _get_swatch(range(232 + 12, 232 + 12 + 6))
        lines.append("    ".join(["", swatch1, swatch2]))
        swatch1 = _get_swatch(range(238, 238 + 6))
        swatch2 = _get_swatch(range(238 + 12, 256))
        lines.append("    ".join(["", swatch1, swatch2]))
        lines.append("")

        print("\n".join(lines))
        self.test_attributes()

        print("\n")



from builtins import object

import platform

_is_windows = "Windows" in platform.system()

ERROR_INVALID_PARAMETER = 0x0057
ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004

if _is_windows:
    import os
    import msvcrt
    import ctypes

    from ctypes import wintypes

    kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)

    def _check_bool(result, func, args):
        if not result:
            raise ctypes.WinError(ctypes.get_last_error())
        return args

    LPDWORD = ctypes.POINTER(wintypes.DWORD)
    kernel32.GetConsoleMode.errcheck = _check_bool
    kernel32.GetConsoleMode.argtypes = (wintypes.HANDLE, LPDWORD)
    kernel32.SetConsoleMode.errcheck = _check_bool
    kernel32.SetConsoleMode.argtypes = (wintypes.HANDLE, wintypes.DWORD)


class Win10Mixin(object):
    def __init__(self):
        self.supported_attrs = []
        self.old_mode = None

    def setup(self, appstate):
        global _is_windows
        if not _is_windows:
            return

        # don't assume StandardOutput is a console.
        # open CONOUT$ instead
        fdout = os.open("CONOUT$", os.O_RDWR)
        try:
            hout = msvcrt.get_osfhandle(fdout)
            self.old_mode = wintypes.DWORD()
            kernel32.GetConsoleMode(hout, ctypes.byref(self.old_mode))
            mode = self.old_mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING
            kernel32.SetConsoleMode(hout, mode)
        except OSError as e:
            self._handle_oserror(e)
            raise e
        finally:
            os.close(fdout)

    def teardown(self, appstate):
        global _is_windows
        if not _is_windows:
            return

        # don't assume StandardOutput is a console.
        # open CONOUT$ instead
        fdout = os.open("CONOUT$", os.O_RDWR)
        try:
            hout = msvcrt.get_osfhandle(fdout)
            kernel32.SetConsoleMode(hout, self.old_mode)
        except OSError as e:
            self._handle_oserror(e)
            raise e
        finally:
            os.close(fdout)

    def _handle_oserror(self, e):
        try:
            if e.winerror == ERROR_INVALID_PARAMETER:
                raise NotImplementedError
        except Exception:
            return



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from builtins import map

from .grow import grow
from ..objects.lilyblock import block


NONE = 0
THIN = 1


# These key characters are sorted to help
# the API take any input sequence.
# t = top
# b = bottom
# l = left
# r = right
MAGIC_MAP = {
    "lr": (0, 0, 1, 1),
    "bt": (1, 1, 0, 0),
    "lt": (1, 0, 1, 0),
    "rt": (1, 0, 0, 1),
    "bl": (0, 1, 1, 0),
    "br": (0, 1, 0, 1),
    "blt": (1, 1, 1, 0),
    "brt": (1, 1, 0, 1),
    "lrt": (1, 0, 1, 1),
    "blr": (0, 1, 1, 1),
    "blrt": (1, 1, 1, 1),
}


BOUNDS_MAP = {
    (0, 0, 0, 0): " ",
    (1, 1, 0, 0): "│",
    (0, 0, 1, 1): "─",
    (1, 0, 1, 0): "┘",
    (1, 0, 0, 1): "└",
    (0, 1, 1, 0): "┐",
    (0, 1, 0, 1): "┌",
    (1, 1, 1, 0): "┤",
    (1, 1, 0, 1): "├",
    (1, 0, 1, 1): "┴",
    (0, 1, 1, 1): "┬",
    (1, 1, 1, 1): "┼",
}

STYLE_MAP = {"thin": THIN}


def _getbounds(vectors, style_int):
    vector_key = "".join(sorted(vectors))
    bounds = MAGIC_MAP[vector_key]
    return tuple(map(lambda b: b * style_int, bounds))


def _frombounds(bounds):
    return BOUNDS_MAP[bounds]


def border_char(vectors, style="thin"):
    style_int = STYLE_MAP[style]
    return _frombounds(_getbounds(vectors, style_int))


def bordered(s, bordercolor="", borderstyle="thin"):
    blc = block(s)
    horiz_char = border_char("lr", borderstyle)
    horiz = horiz_char * blc.width
    horiz = grow(horiz, bordercolor)
    top = block(horiz)
    partial = top.append(blc).append(horiz)
    top_left = _brderchar("br", bordercolor, borderstyle)
    top_right = _brderchar("bl", bordercolor, borderstyle)
    bot_left = _brderchar("tr", bordercolor, borderstyle)
    bot_right = _brderchar("tl", bordercolor, borderstyle)
    vert = _brderchar("tb", bordercolor, borderstyle)
    vert = [vert] * blc.height
    l_vert = block([top_left] + vert + [bot_left])
    r_vert = block([top_right] + vert + [bot_right])
    return l_vert + partial + r_vert


def _brderchar(vector, color, style):
    ch = border_char(vector, style)
    return grow(ch, color)



from .state_manager import (
    get_compiler,
    teardown,
    nocolor,
    compile_all,
    colorama,
    CustomTerminal,
)

__all__ = [
    "get_compiler",
    "teardown",
    "nocolor",
    "compile_all",
    "colorama",
    "CustomTerminal",
]



import json

try:
    import importlib.resources as importlib_resources
except ImportError:
    import importlib_resources
from .colors import Color


class InvalidHexCodeError(Exception):
    pass


def get_color(name):
    return palette.get(name)


def rgb(r, g, b):
    return Color(rgb=(r, g, b))


def hsl(h, s, l):
    return Color(hsl=(h, s, l))


def hex_color(hex_code):
    if hex_code[0] == "#":
        hex_code = hex_code[1:]
    if len(hex_code) == 6:
        r_hex = hex_code[0:2]
        g_hex = hex_code[2:4]
        b_hex = hex_code[4:6]
    elif len(hex_code) == 3:
        r_hex = hex_code[0] * 2
        g_hex = hex_code[1] * 2
        b_hex = hex_code[2] * 2
    else:
        raise InvalidHexCodeError(hex_code)
    return rgb(*[int(h, 16) for h in (r_hex, g_hex, b_hex)])


def _load_palette(file_handle):
    loaded_file = json.load(file_handle)
    return {
        color_name: hex_color(hex_code)
        for color_name, hex_code in loaded_file.items()
    }


def _load_palette_internal(palette_file):
    with importlib_resources.path(
        "lilies.style.palettes", palette_file
    ) as path:
        with open(str(path), "r") as f:
            return _load_palette(f)


palette = _load_palette_internal("default.json")


BLACK = rgb(0, 0, 0)
MAROON = rgb(128, 0, 0)
GREEN = rgb(0, 128, 0)
OLIVE = rgb(128, 128, 0)
NAVY = rgb(0, 0, 128)
PURPLE = rgb(128, 0, 128)
TEAL = rgb(0, 128, 128)
SILVER = rgb(192, 192, 192)
DARKGRAY = rgb(128, 128, 128)
RED = rgb(255, 0, 0)
LIME = rgb(0, 255, 0)
YELLOW = rgb(255, 255, 0)
BLUE = rgb(0, 0, 255)
MAGENTA = rgb(255, 0, 255)
CYAN = rgb(0, 255, 255)
WHITE = rgb(255, 255, 255)


ORDERED_ANSI_16_COLOR_PALETTE = [
    BLACK,
    MAROON,
    GREEN,
    OLIVE,
    NAVY,
    PURPLE,
    TEAL,
    SILVER,
    DARKGRAY,
    RED,
    LIME,
    YELLOW,
    BLUE,
    MAGENTA,
    CYAN,
    WHITE,
]


ANSI_16_COLOR_PALETTE = set(ORDERED_ANSI_16_COLOR_PALETTE)



from .colors import Color


class StyleDiff(object):
    def __init__(self, fg=None, bg=None, attrs={}, full_reset=False):
        self.fg = fg
        self.bg = bg
        self.attrs = attrs

        # Full reset overrides all other style changes in favor of
        # a complete reset to defaults. This helps favor the full
        # ANSI reset character, \033[0m, which is cleaner on average.
        self._full_reset = full_reset

    def isreset(self):
        return self._full_reset


class Style(object):
    def __init__(self, fg=Color(), bg=Color(), attrs=[]):
        self.fg = fg
        self.bg = bg
        self.attrs = list(attrs)

    def diff(self, previous):
        if self.is_default() and not previous.is_default():
            return StyleDiff(full_reset=True)
        fg_diff = None
        bg_diff = None
        if self.fg != previous.fg:
            fg_diff = self.fg
        if self.bg != previous.bg:
            bg_diff = self.bg
        these_attrs = set(self.attrs)
        those_attrs = set(previous.attrs)
        new_attrs = these_attrs - those_attrs
        removed_attrs = those_attrs - these_attrs
        attrs_diff = {}
        for attr in new_attrs:
            attrs_diff[attr] = True
        for attr in removed_attrs:
            attrs_diff[attr] = False
        return StyleDiff(fg_diff, bg_diff, attrs_diff)

    def is_default(self):
        return self.fg == Color() and self.bg == Color() and self.attrs == []

    def __hash__(self):
        components = (hash(self.fg), hash(self.bg), tuple(self.attrs))
        return hash(components)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)



from builtins import str
from future.utils import string_types
from .objects.base import LilyBase


def wilt(s):
    if isinstance(s, LilyBase):
        return s.wilt()
    else:
        return str(s)


def isstringish(obj):
    if issubclass(type(obj), LilyBase):
        # This is a hack, since for whatever reason
        # python considers LilyBase and LilyString to
        # both be instances of one another.
        return obj._isstringish()
    return isinstance(obj, (string_types,))


def assert_stringish(obj):
    if not isstringish(obj):
        raise TypeError("Expected something string-like: " + repr(obj))


def islilyblock(obj):
    return isinstance(obj, LilyBase) and obj._isblockish()


def assert_lilyblock(obj):
    if not islilyblock(obj):
        raise TypeError("Expected LilyBlock: " + repr(obj))



from __future__ import unicode_literals

# Leading control character
CSI = "\033["

FULLRESET = 0
BOLD = BRIGHT = 1
DIM = 2
ITALIC = 3
UNDERLINE = 4
BLINK = 5

# Unsupported
################
# RAPIDBLINK = 6
# REVERSE = 7
# CONCEAL = 8

STRIKE = 9

# Unsupported
################
# PRIMARY_FONT = 10
# ALTFONT1 = 11
# ALTFONT2 = 12
# ALTFONT3 = 13
# ALTFONT4 = 14
# ALTFONT5 = 15
# ALTFONT6 = 16
# ALTFONT7 = 17
# ALTFONT8 = 18
# ALTFONT9 = 19
# FRAKTUR = 20
# DOUBLEUNDERLINE = 21

NOBOLDDIM = 22
NOITALIC = 23
NOUNDERLINE = 24
NOBLINK = 25

# Unsupported
################
# 26 is missing?
# NOREVERSE = 27
# NOCONCEAL = 28

NOSTRIKE = 29

# COLORS!
BLACK = 30
RED = 31
GREEN = 32
YELLOW = 33
BLUE = 34
MAGENTA = 35
CYAN = 36
LIGHTGRAY = 37

NOCOLOR = 39


# 16-color extended,
# Only kind of supported
DARKGRAY = 90
BRIGHTRED = 91
BRIGHTGREEN = 92
BRIGHTYELLOW = 93
BRIGHTBLUE = 94
BRIGHTMAGENTA = 95
BRIGHTCYAN = 96
WHITE = 97


ATTR_ON_CODES = {
    "bold": BOLD,
    "dim": DIM,
    "italic": ITALIC,
    "underline": UNDERLINE,
    "blink": BLINK,
}


ATTR_OFF_CODES = {
    "bold": NOBOLDDIM,
    "dim": NOBOLDDIM,
    "italic": NOITALIC,
    "underline": NOUNDERLINE,
    "blink": NOBLINK,
}


def fg_to_bg(ansi):
    if ansi is None:
        return None
    return ansi + 10


def esc(code):
    if code is None:
        return ""
    return "{csi}{code}m".format(csi=CSI, code=code)



from builtins import super

from .ansi8open import Ansi8OpenTerminal
from .exceptions import UnsupportedColorException
from . import ansicodes
from ..style import Style
from ..style.palette import (
    BLACK,
    MAROON,
    GREEN,
    OLIVE,
    NAVY,
    PURPLE,
    TEAL,
    SILVER,
    DARKGRAY,
    RED,
    LIME,
    YELLOW,
    BLUE,
    MAGENTA,
    CYAN,
    WHITE,
)

DARK_TO_LIGHT = {
    BLACK: DARKGRAY,
    MAROON: RED,
    GREEN: LIME,
    OLIVE: YELLOW,
    NAVY: BLUE,
    PURPLE: MAGENTA,
    TEAL: CYAN,
    SILVER: WHITE,
}


class Ansi16Terminal(Ansi8OpenTerminal):
    def __init__(self):
        super().__init__()
        # new bright colors
        self.supported_colors += [
            DARKGRAY,
            RED,
            LIME,
            YELLOW,
            BLUE,
            MAGENTA,
            CYAN,
            WHITE,
        ]
        self.fg_colormap.update(
            {
                DARKGRAY: [ansicodes.BLACK, ansicodes.BOLD],
                RED: [ansicodes.RED, ansicodes.BOLD],
                LIME: [ansicodes.GREEN, ansicodes.BOLD],
                YELLOW: [ansicodes.YELLOW, ansicodes.BOLD],
                BLUE: [ansicodes.BLUE, ansicodes.BOLD],
                MAGENTA: [ansicodes.MAGENTA, ansicodes.BOLD],
                CYAN: [ansicodes.CYAN, ansicodes.BOLD],
                WHITE: [ansicodes.WHITE, ansicodes.BOLD],
            }
        )
        self.supported_attrs = ["italic", "dim", "underline", "blink"]

    def configure_style(self, style):
        configured = super().configure_style(style)
        if style.fg.isreset():
            return configured
        configured.fg = self._hsl_to_16color(*style.fg.hsl)
        return configured

    def assert_compatible_stylediff(self, style_diff):
        super().assert_compatible_stylediff(style_diff)
        if style_diff.bg is not None:
            self.assert_compatible_bgcolor(style_diff.bg)

    def assert_compatible_bgcolor(self, color):
        # check if it's in the DARK colors
        # light backgrounds aren't supported.
        if color not in DARK_TO_LIGHT:
            raise UnsupportedColorException("Unsupported background color")

    def _hsl_to_16color(self, h, s, l):
        color = super()._hsl_to_8color(h, s, l)
        # For grayscale, we want to optimize for a different lightness
        if color.hsl[1] == 0:
            if l < 25:
                return BLACK
            elif l < 65:
                return DARKGRAY
            elif l < 85:
                return SILVER
            else:
                return WHITE
        if l > 35:
            return DARK_TO_LIGHT[color]
        return color

    def test(self):
        # remove the first "No color" color and unsupported backgrounds
        fg_colors = self.supported_colors[1:]
        bg_colors = self.supported_colors[1:9]
        swatches = [
            ["" for x in range(len(fg_colors))] for y in range(len(bg_colors))
        ]

        for i in range(len(fg_colors)):
            for j in range(len(bg_colors)):
                style = Style(fg=fg_colors[i], bg=bg_colors[j])
                diff = style.diff(Style())
                reset_diff = Style().diff(style)
                color = self.encode_sequence(diff)
                reset = self.encode_sequence(reset_diff)
                swatches[j][i] = "{clr} A {reset}".format(
                    clr=color, reset=reset
                )

        text_rows = ["~Lilies~", "Terminal test: ANSI 16 Colors", ""]
        text_rows.append("Color table:")
        color_rows = map(lambda rowlist: "".join(rowlist), swatches)

        print("\n".join(text_rows + list(color_rows)))
        self.test_attributes()
        print("\n")



import os
from future.utils import string_types
from ..objects.base import LilyBase
from ..objects.lilyblock import LilyBlock
from ..objects.lilystring import LilyString
from ..style import parse_style


def grow(s, style=None, *args, **kwargs):
    if isinstance(s, LilyBase):
        if style is None:
            return s
        else:
            return grow(s.wilt(), style, *args, **kwargs)
    style = parse_style(style)
    if isinstance(s, (string_types,)):
        if os.linesep in s:
            return LilyBlock(s, style, *args, **kwargs)
        return LilyString(s, style, *args, **kwargs)
    elif hasattr(s, "__iter__"):
        if len(s) == 1:
            return LilyString(s[0], style, *args, **kwargs)
        return LilyBlock(s, style, *args, **kwargs)
    else:
        return LilyString(s, style, *args, **kwargs)



from future.utils import string_types
from . import colors
from .palette import get_color
from .styles import Style


class InvalidStyleError(Exception):
    pass


ATTRIBUTE_NAME_MAP = {
    "bold": "bold",
    "strong": "bold",
    "bright": "bold",
    "dim": "dim",
    "italic": "italic",
    "strike": "strike",
    "strikethrough": "strike",
    "struck": "strike",
    "struckthrough": "strike",
    "underlined": "underline",
    "underline": "underline",
    "underscore": "underline",
    "underscored": "underline",
    "blink": "blink",
    "flash": "blink",
}


def _strip_all(itr):
    return list(map(lambda s: s.strip(), itr))


def _is_attr(name):
    return name in ATTRIBUTE_NAME_MAP


def _translate_attr(attr):
    return ATTRIBUTE_NAME_MAP[attr]


def parse_style(style):
    if style is None:
        return Style()

    return _parse_style_str(style)


def _parse_style_str(style):
    """
    parses style as a style string.
    Format:
    comma, separated, attributes FG_COLOR on BG_COLOR

    Examples:

    Bold, underlined, red text on a black background
    bold, underlined red on black

    Red text on blue background:
    red on blue

    Green text:
    green

    Normal text on a white background:
    on white
    """
    if not isinstance(style, (string_types,)):
        raise InvalidStyleError("Expected a string type")
    # Step 1: split out attributes with commas
    # and clean components for whitespace
    step1 = style.split(",")
    step1 = _strip_all(step1)

    # at this point, we have all but the last attribute
    # parted out, with the rest of we need to parse on
    # the right.
    attrs = step1[:-1]
    step1 = step1[-1]
    for attr in attrs:
        if not _is_attr(attr):
            raise InvalidStyleError(
                "Expected attribute: {attr}".format(attr=attr)
            )

    # Step 2: Let's try to pull the last attribute out
    # by splitting by word, and testing the first word.
    # If it's not an attribute, we should have no attributes.
    # Otherwise, someone input something like this:
    # bold, on green
    step2 = step1.split()
    step2 = _strip_all(step2)
    has_attr = len(step2) > 0 and _is_attr(step2[0])
    expect_no_attributes = len(step2) > 1 and not has_attr
    if expect_no_attributes and len(attrs) > 0:
        raise InvalidStyleError("Unexpected comma")
    if has_attr:
        attrs.append(step2[0])
        # remove the attribute from our set, we're done with it
        step2 = step2[1:]
    # we allow for some aliases, so this will unalias them
    attrs = list(map(_translate_attr, attrs))

    # Step 3: We need to part out a background color if
    # it exists. This is entirely dependent on the 'on' keyword.
    # Reminder that step2 at this point is still a list.
    if "on" in step2:
        ix = step2.index("on")
        bg_str = "".join(step2[ix + 1 :])
        bg_str = _clean_colorname(bg_str)
        bg = get_color(bg_str)
        if not bg:
            raise InvalidStyleError(
                "Invalid background color: {bg}".format(bg=bg_str)
            )
        # Step 3 should now be JUST a foreground color.
        step4 = "".join(step2[:ix])
    else:
        # Use default color
        bg = colors.Color()
        step4 = "".join(step2)

    # Step 4:
    # All that should be left at this point is the fg color.
    # Since we allow empty color names to be 'default',
    # account for that as well.
    fg_str = _clean_colorname(step4)
    if len(fg_str) == 0:
        fg = colors.Color()
    else:
        fg = get_color(fg_str)
        if not fg:
            raise InvalidStyleError(
                "Invalid foreground color: {fg}".format(fg=fg_str)
            )

    # All done
    return Style(fg, bg, attrs)


def _clean_colorname(name):
    # We accept camelCase, snake_case,
    # spaces, whatever formatting people like.

    if len(name) > 30:
        return ""
    name = name.strip()
    unsnaked = "".join(name.split("_"))
    condensed = "".join(unsnaked.split())
    return condensed.lower()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
import os
from lilies import grow, wilt, bordered


class TestBordered(unittest.TestCase):
    def setUp(self):
        self.string = "hey"
        self.lilystring = grow("hey", "red")
        self.lilyblock = grow(["quick", "brown", "fox"], "blue")

    def test_bordered_python_string(self):
        control = os.linesep.join(["┌───┐", "│hey│", "└───┘"])
        test = bordered(self.string)
        self.assertEqual(control, wilt(test))

    def test_bordered_lilystring(self):
        control = os.linesep.join(["┌───┐", "│hey│", "└───┘"])
        test = bordered(self.lilystring)
        self.assertEqual(control, wilt(test))

    def test_bordered_lilyblock(self):
        control = os.linesep.join(
            ["┌─────┐", "│quick│", "│brown│", "│fox  │", "└─────┘"]
        )
        test = bordered(self.lilyblock)
        self.assertEqual(control, wilt(test))



from builtins import object
import colorsys


def rgb_to_hsl(r, g, b):
    r_ratio = r / 255.0
    g_ratio = g / 255.0
    b_ratio = b / 255.0

    hsl_ratios = colorsys.rgb_to_hls(r_ratio, g_ratio, b_ratio)
    return (hsl_ratios[0] * 360, hsl_ratios[2] * 100, hsl_ratios[1] * 100)


def hsl_to_rgb(h, s, l):
    h_ratio = h / 360.0
    l_ratio = l / 100.0
    s_ratio = s / 100.0

    rgb_ratios = colorsys.hls_to_rgb(h_ratio, l_ratio, s_ratio)
    return tuple(map(lambda n: int(round(n * 255)), rgb_ratios))


class Color(object):
    def __init__(self, rgb=None, hsl=None):
        if rgb is None and hsl is None:
            # This is considered "NO COLOR"
            # and is the default console color.
            self.rgb = None
            self.hsl = None
            return
        if rgb is None:
            rgb = hsl_to_rgb(*hsl)
        if hsl is None:
            hsl = rgb_to_hsl(*rgb)
        self.rgb = rgb
        self.hsl = hsl

    def isreset(self):
        return self.rgb is None and self.hsl is None

    def __hash__(self):
        # This is a hack to make sure that None is not the same
        # as an undefined color.
        if self.rgb is None:
            return hash("default")
        return hash(self.rgb)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)



from builtins import object
from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass

from . import ansicodes


class BaseTerminal(with_metaclass(ABCMeta, object)):
    @abstractmethod
    def setup(self, appstate):
        raise NotImplementedError("setup")

    @abstractmethod
    def teardown(self, appstate):
        raise NotImplementedError("teardown")

    def assert_compatible_stylediff(self, style_diff):
        if style_diff.fg is not None:
            self.assert_compatible_color(style_diff.fg)
        if style_diff.bg is not None:
            self.assert_compatible_color(style_diff.bg)
        self.assert_compatible_attrs(style_diff.attrs.keys())

    @abstractmethod
    def configure_style(self, style):
        raise NotImplementedError("configure_style")

    @abstractmethod
    def assert_compatible_color(self, color):
        raise NotImplementedError("assert_compatible_color")

    @abstractmethod
    def assert_compatible_attrs(self, attrs):
        raise NotImplementedError("assert_compatible_attrs")

    @abstractmethod
    def _build_fg_codes(self, fg_color):
        raise NotImplementedError("_build_fg_codes")

    @abstractmethod
    def _build_bg_codes(self, bg_color):
        raise NotImplementedError("_build_bg_codes")

    @abstractmethod
    def _build_attr_codes(self, attrs):
        raise NotImplementedError("_build_attr_codes")

    @abstractmethod
    def test(self):
        raise NotImplementedError("test")

    @abstractmethod
    def _build_reset_sequence(self):
        raise NotImplementedError("_build_reset_sequence")

    def encode_sequence(self, style_diff):
        self.assert_compatible_stylediff(style_diff)
        if style_diff.isreset():
            return "".join(self._build_reset_sequence())
        fg_codes = self._build_fg_codes(style_diff.fg)
        bg_codes = self._build_bg_codes(style_diff.bg)
        attr_codes = self._build_attr_codes(style_diff.attrs)
        fg_seq = "".join(map(ansicodes.esc, fg_codes))
        bg_seq = "".join(map(ansicodes.esc, bg_codes))
        attr_chars = map(ansicodes.esc, attr_codes)
        attr_seq = "".join(attr_chars)
        return attr_seq + fg_seq + bg_seq



from .win10mixin import Win10Mixin
from .win32mixin import Win32Mixin
from .truecolor import TrueColorTerminal
from .ansi256 import Ansi256Terminal
from .aixterm import AixTerminal


class WinTrueColor(Win10Mixin, TrueColorTerminal):
    pass


class Win256Color(Win10Mixin, Ansi256Terminal):
    pass


class WinLegacy(Win32Mixin, AixTerminal):
    pass



import os
import re
import setuptools


def read_file(path):
    with open(os.path.join(os.path.dirname(__file__), path)) as fp:
        return fp.read()


def _get_version_match(content):
    # Scan package __init__.py for a __version__ line
    # and grab the output literal
    regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version_match = re.search(regex, content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def get_version(path):
    return _get_version_match(read_file(path))


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setuptools.setup(
    name="lilies",
    version=get_version(os.path.join("lilies", "__init__.py")),
    description="A tool for creating colorful, formatted command line output",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mrz1988/lilies",
    license="MIT",
    author="Matt Zychowski",
    author_email="mrz2004@gmail.com",
    packages=setuptools.find_packages(),
    package_data={"lilies.style.palettes": ["*.json"]},
    install_requires=[
        "colorama",
        "future",
        'importlib_resources; python_version < "3.7"',
    ],
    python_requires=">=3.5",
)



class UnsupportedStyleException(Exception):
    """
    Base exception when an unsupported style
    is encountered
    """

    pass


class UnsupportedColorException(UnsupportedStyleException):
    """
    Raised when a color has not been properly translated
    for the terminal before attempting rendering
    """

    pass


class UnsupportedAttributeException(UnsupportedStyleException):
    """
    Raised when an attribute is not supported by
    the current terminal but rendering was attempted
    """

    pass



from . import ansicodes
from ..style import Style


class AnsiAttributesMixin(object):
    """
    Adds functionality to detect and encode standard
    ansi style attributes, must also extend Ansi8Terminal
    or a subclass for this to work.
    """

    def _build_attr_codes(self, attrs):
        on_codes = []
        off_codes = []
        bold_dim_off = False
        # Bold and Dim have the same off char.
        # This char should come FIRST, and then
        # other chars can go after to build up
        # the rest of the sequence.
        for attr in attrs:
            if attrs[attr]:
                on_codes.append(self._attr_on_code(attr))
            else:
                if attr not in ["bold", "dim"]:
                    off_codes.append(self._attr_off_code(attr))
                elif not bold_dim_off:
                    # Bold and Dim have the same off char.
                    # Make sure we only add one.
                    off_codes.append(self._attr_off_code(attr))
                    bold_dim_off = True
        return off_codes + on_codes

    def _attr_on_code(self, attr):
        return ansicodes.ATTR_ON_CODES[attr]

    def _attr_off_code(self, attr):
        return ansicodes.ATTR_OFF_CODES[attr]

    def test_attributes(self):
        if len(self.supported_attrs) == 0:
            return

        lines = ["Supported styles:"]
        attr_strs = []
        no_style = Style(attrs=[])
        for attr in self.supported_attrs:
            cur_style = Style(attrs=[attr])
            diff = cur_style.diff(no_style)
            ansi = self.encode_sequence(diff)
            attr_strs.append(
                "{ansi}{attr}{reset}".format(
                    ansi=ansi, attr=attr, reset="\033[0m"
                )
            )
        lines.append("  ".join(attr_strs))
        print("\n".join(lines))



from builtins import super

from ..style import Style
from .ansi8 import Ansi8Terminal
from .attr_mixin import AnsiAttributesMixin


class Ansi8OpenTerminal(Ansi8Terminal, AnsiAttributesMixin):
    def __init__(self):
        super().__init__()
        self.supported_attrs = ["bold", "dim", "underline"]

    def test(self):
        # remove the first "No color" color
        colors = self.supported_colors[1:]
        swatches = [
            ["" for x in range(len(colors))] for y in range(len(colors))
        ]

        for i in range(len(colors)):
            for j in range(len(colors)):
                style = Style(fg=colors[i], bg=colors[j])
                diff = style.diff(Style())
                reset_diff = Style().diff(style)
                color = self.encode_sequence(diff)
                reset = self.encode_sequence(reset_diff)
                swatches[i][j] = "{clr} A {reset}".format(
                    clr=color, reset=reset
                )

        text_rows = ["~Lilies~", "Terminal test: ANSI 8 Colors, Extended", ""]
        text_rows.append("Color table:")
        color_rows = map(lambda rowlist: "".join(rowlist), swatches)

        print("\n".join(text_rows + list(color_rows)))
        self.test_attributes()
        print("\n")



from abc import ABC, abstractmethod


class LilyBase(ABC):
    @abstractmethod
    def wilt(self):
        raise NotImplementedError("wilt")

    @abstractmethod
    def _isstringish(self):
        raise NotImplementedError("_isstringish")

    @abstractmethod
    def _isblockish(self):
        raise NotImplementedError("_isblockish")

    @property
    @abstractmethod
    def width(self):
        raise NotImplementedError("width")

    @property
    @abstractmethod
    def height(self):
        raise NotImplementedError("height")

    @classmethod
    def __subclasshook__(cls, subclass):
        required = ["wilt", "_isstringish", "_isblockish"]
        for r in required:
            if not any(r in c.__dict__ for c in subclass.__mro__):
                return NotImplemented
        return True



from copy import deepcopy
from ..objects.lilyblock import block
from ..objects.lilystring import LilyString
from ..style import Style
from .grow import grow


class Resizer:
    def __init__(self, width, height, align, add_elipsis):
        self.width = width
        self.height = height
        self.align_x, self.align_y = self._parse_align(align)
        self.add_elipsis = add_elipsis

    def resize(self, stringlike, character=" "):
        lilyblock = block(grow(stringlike))
        if self.height == -1:
            height = lilyblock.height
        else:
            height = self.height
        lilyblock = self._resize_vertical(lilyblock, height, character)

        if self.width == -1:
            width = lilyblock.width
        else:
            width = self.width
        resized_rows = [
            self._resize_row(row, width, character) for row in lilyblock
        ]

        return grow(resized_rows)

    def _parse_align(self, align):
        if align == "center":
            return "center", "center"
        if align in ("left", "right"):
            return align, "top"
        if align in ("top", "bottom"):
            return "left", align
        align_components = align.split()
        if len(align_components) != 2:
            raise TypeError("Invalid align value: " + str(align))
        align_y, align_x = align_components
        if align_x not in ("left", "right", "center"):
            raise TypeError("Invalid horizontal align value: " + str(align_x))
        if align_y not in ("top", "bottom", "center"):
            raise TypeError("Invalid vertical align value: " + str(align_y))
        return align_x, align_y

    def _resize_row(self, lilystring, width, character):
        if lilystring.width <= width:
            return self._widen(lilystring, width, character)
        else:
            return self._truncate(lilystring, width)

    def _widen(self, lilystring, width, character):
        if len(lilystring) == width:
            return deepcopy(lilystring)

        if self.align_x == "center":
            return lilystring.center(width, character)
        elif self.align_x == "right":
            return lilystring.rjust(width, character)
        else:
            return lilystring.ljust(width, character)

    def _truncate(self, lilystring, width):
        if self.add_elipsis:
            elipsis = LilyString("..", lilystring.style_at(-1))
            if self.width <= len(elipsis):
                # no elipsis if it'd be too small
                return lilystring[:width]
            truncated = lilystring[: width - len(elipsis)]
            return truncated + elipsis
        return lilystring[:width]

    def _resize_vertical(self, lilyblock, height, character):
        if self.align_y == "center":
            return self._resize_vertical_centered(lilyblock, height, character)
        elif self.align_y == "bottom":
            return self._resize_vertical_uncentered(
                lilyblock, height, character, bottom=True
            )
        else:
            return self._resize_vertical_uncentered(
                lilyblock, height, character, bottom=False
            )

    def _resize_vertical_centered(self, lilyblock, height, character):
        height_difference = height - lilyblock.height

        # We want the top half to always have equal or less change,
        # Expanding or contracting by an odd number will cause the
        # Lower rows to be affected more than the top rows.
        # Python will always round remainders toward negative infinity.
        # Since we want it rounded towards 0, the positive remainder 1,
        # if it exists, should get added with the quotient.
        if height_difference > 0:
            top_difference = height_difference // 2
        else:
            top_difference = sum(divmod(height_difference, 2))

        resized_top_only = self._resize_vertical_uncentered(
            lilyblock,
            lilyblock.height + top_difference,
            character,
            bottom=True,
        )
        return self._resize_vertical_uncentered(
            resized_top_only, height, character
        )

    def _resize_vertical_uncentered(
        self, lilyblock, height, character, bottom=False
    ):
        height_difference = height - lilyblock.height
        if height_difference < 0:
            if bottom:
                return lilyblock[abs(height_difference) :]
            else:
                return lilyblock[:height]
        try:
            style = character.style_at(0)
        except AttributeError:
            if lilyblock.height:
                if bottom:
                    style = lilyblock._rows[0].style_at(0)
                else:
                    style = lilyblock._rows[-1].style_at(0)
            else:
                style = Style()
        if character is None:
            character = " "
        fill = block([LilyString(character, style)] * height_difference)
        if bottom:
            return fill.append(lilyblock)
        else:
            return lilyblock.append(fill)


def resize(
    stringlike,
    width=-1,
    height=-1,
    align="left",
    character=None,
    add_elipsis=True,
):
    resizer = Resizer(width, height, align, add_elipsis=add_elipsis)
    return resizer.resize(stringlike, character)



from .nocolor import NoColorTerminal
from . import ansicodes
from .exceptions import UnsupportedColorException
from ..style import Color, Style
from ..style.palette import (
    BLACK,
    MAROON,
    GREEN,
    OLIVE,
    NAVY,
    PURPLE,
    TEAL,
    SILVER,
)


class Ansi8Terminal(NoColorTerminal):
    """
    An 8-color ansi terminal
    """

    def __init__(self):
        self.supported_colors = [
            Color(),
            BLACK,
            MAROON,
            GREEN,
            OLIVE,
            NAVY,
            PURPLE,
            TEAL,
            SILVER,
        ]
        self.supported_attrs = []
        self.fg_colormap = {
            None: [],
            Color(): [ansicodes.NOCOLOR],
            BLACK: [ansicodes.BLACK],
            MAROON: [ansicodes.RED],
            GREEN: [ansicodes.GREEN],
            OLIVE: [ansicodes.YELLOW],
            NAVY: [ansicodes.BLUE],
            PURPLE: [ansicodes.MAGENTA],
            TEAL: [ansicodes.CYAN],
            SILVER: [ansicodes.LIGHTGRAY],
        }
        self.bg_colormap = {}
        for key in self.fg_colormap:
            val = self.fg_colormap[key]
            bgcodes = map(ansicodes.fg_to_bg, val)
            self.bg_colormap[key] = bgcodes

    def configure_style(self, style):
        """
        Configure the style for this terminal
        """
        if style.fg.isreset():
            fg = Color()
        else:
            fg = self._hsl_to_8color(*style.fg.hsl)
        if style.bg.isreset():
            bg = Color()
        else:
            bg = self._hsl_to_8color(*style.bg.hsl)

        def issupported(style):
            return style in self.supported_attrs

        attrs = filter(issupported, style.attrs)
        return Style(fg, bg, attrs)

    def _build_fg_codes(self, fg_color):
        ansi = self.fg_colormap.get(fg_color)
        if ansi is None:
            raise UnsupportedColorException()
        return ansi

    def _build_bg_codes(self, bg_color):
        return map(ansicodes.fg_to_bg, self._build_fg_codes(bg_color))

    def _build_attr_codes(self, attrs):
        return []

    def _hsl_to_8color(self, h, s, l):
        # grayscale, low saturation
        if s < 10:
            return BLACK if l < 20 else SILVER

        # higher saturation, go by hue
        if h < 30 or h > 345:
            return MAROON
        if h < 65:
            return OLIVE
        if h < 145:
            return GREEN
        if h < 215:
            return TEAL
        if h < 260:
            return NAVY
        return PURPLE

    def _build_reset_sequence(self):
        return [ansicodes.esc(0)]

    def test(self):
        # remove the first "No color" color
        colors = self.supported_colors[1:]
        swatches = [
            ["" for x in range(len(colors))] for y in range(len(colors))
        ]

        for i in range(len(colors)):
            for j in range(len(colors)):
                style = Style(fg=colors[i], bg=colors[j])
                diff = style.diff(Style())
                reset_diff = Style().diff(style)
                color = self.encode_sequence(diff)
                reset = self.encode_sequence(reset_diff)
                swatches[i][j] = "{clr} A {reset}".format(
                    clr=color, reset=reset
                )

        text_rows = ["~Lilies~", "Terminal test: ANSI 8 Colors", ""]
        text_rows.append("Color table:")
        color_rows = map(lambda rowlist: "".join(rowlist), swatches)

        print("\n".join(text_rows + list(color_rows)))



import os
import re

from ..base_utils import wilt
from ..style import parse_style
from ..objects.lilyblock import LilyBlock
from ..objects.lilystring import LilyString
from .grow import grow


class Highlighter:
    def __init__(self, pattern, style_str, flags, num):
        self.pattern = pattern
        self.style = parse_style(style_str)
        self.flags = flags
        self.num = num

    def highlight(self, stringish):
        lilystring = self._convert_to_lilystring(stringish)
        lilystring = self._style_regex(lilystring)
        if os.linesep in lilystring:
            return LilyBlock(lilystring)
        return lilystring

    def _style_regex(self, lilystring):
        raw_text = wilt(lilystring)
        matches = re.finditer(self.pattern, raw_text, self.flags)
        if self.num > 0:
            matches = list(matches)[: self.num]
        for group in matches:
            lilystring = self._style_slice(
                lilystring, group.start(), group.end()
            )
        return lilystring

    def _style_slice(self, lilystring, start, stop):
        left = lilystring[:start]
        middle_text = wilt(lilystring[start:stop])
        middle = LilyString(middle_text, self.style)
        right = lilystring[stop:]
        return left + middle + right

    def _convert_to_lilystring(self, stringish):
        lily = grow(stringish)
        if type(lily) == LilyBlock:
            endl = LilyString(os.linesep)
            lily = endl.join(lily._copy_rows())
        return lily


def highlight(stringlike, pattern, style_str="yellow", flags=0, num=0):
    highlighter = Highlighter(pattern, style_str, flags, num)
    return highlighter.highlight(stringlike)



# -*- coding: utf-8 -*-
from builtins import map
from builtins import str
from builtins import range
from builtins import object
from future.utils import string_types, python_2_unicode_compatible
from copy import deepcopy
import os

from ..base_utils import isstringish, wilt
from ..compiler import get_compiler, compile_all
from ..style import Style
from .base import LilyBase


def lstr(s, *args, **kwargs):
    if isinstance(s, LilyString) and s._isstringish():
        return s
    else:
        return LilyString(s, *args, **kwargs)


@python_2_unicode_compatible
class LilyString(LilyBase):
    def __init__(self, s="", style=Style()):
        self._pieces = []
        self._append(s, style)

    def __str__(self):
        return get_compiler().compile(self._pieces)

    def __repr__(self):
        return "c'" + self.__str__() + "'"

    def __len__(self):
        return sum(map(len, self._pieces))

    def __int__(self):
        return int(self.wilt())

    def __long__(self):
        return int(self.wilt())

    def __float__(self):
        return float(self.wilt())

    def __getitem__(self, key):
        if isinstance(key, slice):
            slice_obj = key
        elif isinstance(key, int):
            slice_obj = slice(key, key + 1, None)
        else:
            raise TypeError("Invalid argument type, looking for int or slice")

        return self._getslice(slice_obj)

    def __iter__(self):
        for p in self._pieces:
            style = p.style
            for c in p.text:
                yield LilyString(c, style)

    def __add__(self, other):
        new_pretty = LilyString()
        try:
            newpieces = self._pieces + other._pieces
            new_pretty._pieces = newpieces
        except AttributeError:
            new_pretty._pieces = deepcopy(self._pieces)
            new_pretty._append(other)
        new_pretty._flatten()
        return new_pretty

    def __radd__(self, other):
        try:
            return other.__add__(self)
        except TypeError:
            return LilyString(other).__add__(self)

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError("can't multiply sequence by non-int")
        if other < 1:
            return LilyString()
        newpieces = self._pieces * other
        new_pretty = LilyString()
        new_pretty._pieces = newpieces
        new_pretty._flatten()
        return new_pretty

    def __rmul__(self, other):
        return self.__mul__(other)

    def __iadd__(self, other):
        return self.__add__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __gt__(self, other):
        return self.wilt() > other

    def __lt__(self, other):
        return self.wilt() < other

    def __ge__(self, other):
        return self.wilt() >= other

    def __le__(self, other):
        return self.wilt() <= other

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        with compile_all:
            return hash(self.__str__())

    def __contains__(self, other):
        if not isstringish(other):
            other = repr(other)
            msg = other + " is not a required stringish object"
            raise TypeError(msg)
        if not isinstance(other, LilyString):
            return other in self.wilt()
        len_self = len(self)
        len_other = len(other)
        if len_other == 0:
            return True
        if len_other > len_self:
            return False
        if len_other == len_self:
            return other == self

        # basic substring search
        for i in range(len_self - len_other):
            for j in range(len_other):
                if self[i + j] != other[j]:
                    break
                if j == len_other - 1:
                    return True
        return False

    def __reversed__(self):
        revd = self[::-1]
        revd._flatten()
        return revd

    def wilt(self):
        sb = u""
        for piece in self._pieces:
            sb += piece.text
        return sb

    @property
    def height(self):
        return 1

    @property
    def width(self):
        return len(self)

    def _isstringish(self):
        return True

    def _isblockish(self):
        return False

    def is_text(self, text):
        return self.wilt() == wilt(text)

    def isnt_text(self, text):
        return self.wilt() != wilt(text)

    def ljust(self, length, character=None):
        if character is None:
            character = LilyString(" ", self.style_at(-1))
        if len(character) != 1:
            raise TypeError(
                "The fill character must be exactly one character long"
            )
        padding_width = length - self.width
        try:
            style = character.style_at(0)
        except AttributeError:
            style = Style()
        padding = LilyString(wilt(character * padding_width), style)
        return deepcopy(self) + padding

    def rjust(self, length, character=None):
        if character is None:
            character = LilyString(" ", self.style_at(0))
        if len(character) != 1:
            raise TypeError(
                "The fill character must be exactly one character long"
            )
        padding_width = length - self.width
        try:
            style = character.style_at(0)
        except AttributeError:
            style = Style()
        padding = LilyString(wilt(character * padding_width), style)
        return padding + deepcopy(self)

    def center(self, length, character=None):
        if character is None:
            character = " "
            left_style = self.style_at(0)
            right_style = self.style_at(-1)
        else:
            try:
                left_style = character.style_at(0)
                right_style = character.style_at(0)
            except AttributeError:
                left_style = Style()
                right_style = Style()
        if len(character) != 1:
            raise TypeError(
                "The fill character must be exactly one character long"
            )

        padding_width = length - self.width
        left_spaces = padding_width // 2
        left = LilyString(wilt(character * left_spaces), left_style)

        right_spaces = left_spaces + (padding_width % 2)
        right = LilyString(wilt(character * right_spaces), right_style)

        return left + deepcopy(self) + right

    def _flatten(self):
        piece_length = len(self._pieces)
        if piece_length == 0:
            return
        newpieces = []
        for i in range(piece_length):
            if len(self._pieces[i]) == 0:
                continue
            if len(newpieces) == 0:
                newpieces.append(deepcopy(self._pieces[i]))
                continue
            if self._pieces[i].style == newpieces[-1].style:
                newpieces[-1].text += self._pieces[i].text
            else:
                newpieces.append(deepcopy(self._pieces[i]))
        self._pieces = newpieces

    def _getslice(self, sliceobj):
        chars = self.wilt()
        ixs = list(range(*sliceobj.indices(len(chars))))
        _new = LilyString()
        for i in ixs:
            _new += LilyString(chars[i], self.style_at(i))
        _new._flatten()
        return _new

    def _append(self, s, style=Style()):
        if s == "":
            return
        self._pieces.append(LilyStringPiece(s, style))

    def join(self, components):
        if not hasattr(components, "__iter__"):
            raise TypeError("can only join an iterable")
        if len(components) == 0:
            return LilyString()
        if len(components) == 1:
            s = components[0]
            if isinstance(s, LilyString):
                return s
            if isinstance(s, (string_types,)):
                return LilyString(s)
            raise TypeError("Was not a stringish type: " + repr(s))
        new_comps = components[0]
        copy = deepcopy(self)
        for i in range(1, len(components)):
            new_comps += copy + components[i]
        return new_comps

    def split(self, sep=None, maxsplit=-1):
        if isinstance(sep, LilyString):
            sep = sep.wilt()
        split_pieces = []
        # Split each string piece individually, storing resulting LilyString
        #   sets in split_pieces
        for piece in self._pieces:
            split_piece = piece.text.split(sep)
            piece_style = piece.style
            if maxsplit != -1:
                # FIXME: bail earlier on this rather than re-splitting
                # everything
                prev_splits = sum([len(p) - 1 for p in split_pieces])
                tot_splits = len(split_piece) + prev_splits - 1
                if tot_splits > maxsplit:
                    split_times = maxsplit - prev_splits
                    split_piece = piece.text.split(sep, split_times)

            split_piece = [LilyString(p, piece_style) for p in split_piece]
            split_pieces.append(split_piece)

        if len(split_pieces) == 0:
            return [] if sep is None else [LilyString("")]

        output = split_pieces[0]
        for i in range(1, len(split_pieces)):
            # join together outside components (for output components that span
            #   multiple pieces)
            output[-1] += split_pieces[i][0]
            # add the rest of the pieces
            output += split_pieces[i][1:]
        return output

    def find(self, s, start=None, end=None):
        s = wilt(s)
        return self.wilt().find(s, start, end)

    def rfind(self, s, start=None, end=None):
        s = wilt(s)
        return self.wilt().rfind(s, start, end)

    def lower(self):
        _new = deepcopy(self)
        for i in range(len(self._pieces)):
            _new._pieces[i].text = _new._pieces[i].text.lower()
        return _new

    def upper(self):
        _new = deepcopy(self)
        for i in range(len(self._pieces)):
            _new._pieces[i].text = _new._pieces[i].text.upper()
        return _new

    def swapcase(self):
        new = deepcopy(self)
        for i in range(len(self._pieces)):
            new._pieces[i].text = new._pieces[i].text.swapcase()
        return new

    def lstrip(self, chars=None):
        new = deepcopy(self)
        if len(new._pieces) == 0:
            return new
        for i in range(len(new._pieces)):
            new._pieces[i].text = new._pieces[i].text.lstrip(chars)
            if len(new._pieces[i]) > 0:
                break
        new._flatten()
        return new

    def rstrip(self, chars=None):
        new = deepcopy(self)
        if len(new._pieces) == 0:
            return new
        for i in reversed(list(range(len(new._pieces)))):
            new._pieces[i].text = new._pieces[i].text.rstrip(chars)
            if len(new._pieces[i]) > 0:
                break
        new._flatten()
        return new

    def strip(self, chars=None):
        return self.lstrip(chars).rstrip(chars)

    def count(self, sub, *args, **kwargs):
        return self.wilt().count(wilt(sub), *args, **kwargs)

    def startswith(self, prefix, *args, **kwargs):
        return self.wilt().startswith(wilt(prefix), *args, **kwargs)

    def endswith(self, suffix, *args, **kwargs):
        return self.wilt().endswith(wilt(suffix), *args, **kwargs)

    def isalnum(self):
        return self.wilt().isalnum()

    def isalpha(self):
        return self.wilt().isalpha()

    def isdigit(self):
        return self.wilt().isdigit()

    def islower(self):
        return self.wilt().islower()

    def isspace(self):
        return self.wilt().isspace()

    def istitle(self):
        return self.wilt().istitle()

    def isupper(self):
        return self.wilt().isupper()

    def style_at(self, index, default=Style()):
        if index < 0:
            index += self.__len__()
        if index >= self.__len__() or index < 0:
            return default
        for p in self._pieces:
            index -= len(p)
            if index < 0:
                return p.style


class LilyStringPiece(object):
    def __init__(self, s, style):
        try:
            self.text = str(s)
        except UnicodeDecodeError:
            self.text = s.decode("utf-8")
        self.style = style

    def __len__(self):
        return len(self.text)


class LilyStringError(Exception):
    pass


class InvalidInputError(Exception):
    pass


endl = LilyString(os.linesep)



from ..terminal import (
    detect_terminal,
    TrueColorTerminal,
    NoColorTerminal,
    WinLegacy,
)
from .compiler import LilyStringCompiler


def get_compiler():
    global _sm
    return _sm.compiler()


def teardown():
    global _sm
    _sm.teardown()


class StateManager(object):
    def __init__(self):
        self.appstate = {}
        self.current_terminal = None
        self.current_compiler = None

    def setup(self):
        if self.current_terminal is None:
            self.setterm(detect_terminal())

    def teardown(self):
        if self.current_terminal is not None:
            self.current_terminal.teardown(self.appstate)
            self.current_terminal = None

    def setterm(self, terminal):
        if self.current_terminal is not None:
            self.current_terminal.teardown(self.appstate)
        self.appstate = {}
        self.current_compiler = None
        terminal.setup(self.appstate)
        self.current_terminal = terminal
        self.current_compiler = LilyStringCompiler(terminal)

    def compiler(self):
        if self.current_compiler is None:
            self.setup()
        return self.current_compiler


class CustomTerminal(object):
    def __init__(self, terminal):
        self.original_term = None
        self.new_term = terminal

    def __enter__(self):
        global _sm
        self.original_term = _sm.current_terminal
        _sm.setterm(self.new_term)
        return self.new_term

    def __exit__(self, t, v, tb):
        global _sm
        if self.original_term is not None:
            _sm.setterm(self.original_term)


_sm = StateManager()


compile_all = CustomTerminal(TrueColorTerminal())
nocolor = CustomTerminal(NoColorTerminal())
colorama = CustomTerminal(WinLegacy())
